/* Idee von N-MODELLBAHN (https://n-modellbahn.de/moba-speed/)

   Version 1.6 ist verbessert von Hardi und Jürgen (fromue)am 28.11.2022

   Stummi Modellbahn-Geschwindigkeit für ARDUINO V1.6
   gemacht u.a. für unseren Stummi-Stammtisch Winnenden
   Arduino Sketch für Modellbahn Geschwindigkeitsmessungen der Maßstäbe H0, TT, N, Spur 0, Spur 1     (Nicht für LGB und 00 geeignet)
   Es können die Mikrocontroller: Arduino Nano.V3, UNO.V3, MEGA eingesetzt werden.

   benötigte Teile:
   2 IR Sensoren mit LM393 Ausgabe
   1 OLED Displays 0.96 und 1.3 Inch mit 128 x 64 Pixel
   Adresse: 0x3C für OLED-Displays von Drittanbietern! Bitte Adresse 0x3D für original Adafruit Produkte nehmen!
   1 Mikrocontroller

   Es können die OLED-Displays Typ 0.96 und 1.3 Inch genommen werden. (Es müssen die entsprechenden Zeilen kommentiert werden!!)
   Die Serielle Ausgabe an z.B. iTrain, Win-Digipet und TrainController haben wir implementiert,
   sowie die Berechnung/ Umrechnung für die KPF-Zeller Speed-Catplus wurde eingefügt.
*/


#define  OLED 96    // Set to 13 to use the 1.3" display

//Entfernung mittig von beiden IR-Sensoren (links/ rechts) - Distanz in mm
const double distance = 290.0;


//Maßstab der Modellbahn 1/xx
int scale = 87;                       // 32 / 45 / 87 / 120 / 160  Nicht für 22.5 / 76.2 da integer Zahl


#define ITRAIN_TEST 0    // Set to 0 to disable 1 = ITrain test with Speed measurement, 2 = print receiver RS232 char and send periodic


//-----------------------------------------
// Digitaler Sensor Eingang links und rechts
#define SENSOR_L 7
#define SENSOR_R 6

////// benötigte Bibliotheken
//#include <SPI.h>                // bei I2C nicht benötigt
#include <Wire.h>                 //I2C Bibliothek
#include <Adafruit_GFX.h>         // Bildschirm-Bibliothek
#define SCREEN_WIDTH 128          // OLED display Breite, in Pixel
#define SCREEN_HEIGHT 64          // OLED display Höhe, in Pixel
#define OLED_RESET 4

#if OLED == 96
   #include <Adafruit_SSD1306.h>     //OLED-Treiber-Bibliothek für 0.96 inch Display
   Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET); // 0.96 INCH OLED Display Initialisierung
#else
   #include <Adafruit_SH1106.h>    //OLED-Treiber-Bibliothek für 1.3 inch Display
   Adafruit_SH1106 display(OLED_RESET);                                    //  1.3 INCH OLED Display Initialisierung
#endif



//Anzeigezeit des Messbildschrims in ms
//int waitingtime = 10000;

// Name der Variablen für Zeit und Geschwindigkeitsmessung
long int deltatime, starttime, freetime_counter;

//Stummi-Logo - Stummilogo für Arduino
const unsigned char PROGMEM logo [] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x22, 0x00, 0x80, 0x00, 0x80, 0x00, 0x04, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x04, 0x41, 0x25, 0x03, 0x03, 0x03, 0x00, 0x04, 0x40,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x03, 0x80, 0x02, 0x00, 0x00, 0x0c, 0x70, 0x01, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x01, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x01, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x02, 0x06, 0x00, 0x03, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x18, 0x01, 0x00, 0x0e, 0x00, 0x02, 0x62, 0x00, 0xc2, 0x02, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x38, 0x00, 0x04, 0x86, 0x00, 0x46, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x01, 0x80, 0x01, 0x00, 0x30, 0x00, 0x0a, 0x4a, 0x18, 0x48, 0x02, 0x10, 0x00,
  0x00, 0x00, 0x00, 0x02, 0x00, 0x03, 0x80, 0x30, 0x00, 0x14, 0x71, 0x58, 0x01, 0x20, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x04, 0x00, 0x0c, 0x80, 0x18, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x08, 0x00, 0x30, 0x00, 0x26, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x08, 0x00, 0x40, 0x00, 0x61, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x0c, 0x00, 0x70, 0x00, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x06, 0x00, 0x0c, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x1c, 0x00, 0x0c, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0xc0, 0x00, 0x0c, 0x38, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0x00, 0x00, 0x17, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x02, 0x00, 0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x04, 0x00, 0x01, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x04, 0x00, 0x0e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x06, 0x00, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x03, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x0f, 0xf8, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x07, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x07, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x07, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x03, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x03, 0xf0, 0x00, 0x00, 0x00, 0x3f, 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x03, 0xf0, 0x00, 0x3f, 0x00, 0x1f, 0x00, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0xe0, 0x00, 0xff, 0x00, 0x0e, 0x2e, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0xe0, 0x00, 0xff, 0x87, 0x8e, 0x3f, 0x1c, 0x00, 0x00, 0xc0, 0x30, 0x00, 0x00,
  0x00, 0x00, 0x01, 0xe3, 0xe0, 0xff, 0x8f, 0x8e, 0x7f, 0x1c, 0x00, 0x01, 0xe8, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0xef, 0xff, 0xff, 0xff, 0xfe, 0x3e, 0x38, 0x00, 0x3f, 0xff, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x3f, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x3e, 0x38, 0x05, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x1e, 0x38, 0x0f, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x01, 0xf8, 0x8f, 0xc1, 0xff, 0x07, 0xfe, 0x3f, 0x7c, 0x1f, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0x78, 0x8d, 0xc9, 0xff, 0x07, 0xff, 0xff, 0xfc, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0xf8, 0x10, 0xc9, 0x0f, 0x3c, 0x20, 0xd0, 0x1c, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0xf0, 0x12, 0x01, 0x0f, 0x01, 0x20, 0x90, 0x1c, 0xff, 0xff, 0xff, 0xc0, 0x70, 0x00,
  0x00, 0x00, 0xf0, 0x16, 0x19, 0x0f, 0x39, 0x2c, 0x94, 0x9c, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0x70, 0x10, 0x00, 0x0e, 0x79, 0x2c, 0x14, 0x9c, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x00, 0x72, 0x90, 0x83, 0x0e, 0x78, 0x4c, 0x24, 0x9c, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00,
  0x00, 0x00, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf9, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00,
  0x00, 0x00, 0x3f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf9, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00,
  0x00, 0x00, 0x01, 0xe7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf1, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00,
  0x00, 0x00, 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x73, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x01, 0xff, 0xf0, 0x3f, 0xc0, 0xff, 0x07, 0xf8, 0x13, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
  0x00, 0x07, 0xff, 0xf8, 0xfd, 0xe3, 0xff, 0x9f, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00,
  0x00, 0x0f, 0xff, 0xf9, 0xf0, 0x7f, 0xc3, 0xfe, 0x3e, 0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00,
  0x00, 0x1f, 0xff, 0xff, 0xe0, 0x7f, 0x81, 0xfc, 0x1e, 0x00, 0x7c, 0x7c, 0x7c, 0x00, 0x00, 0x00,
  0x00, 0x7f, 0xff, 0xf9, 0xe0, 0x7f, 0x01, 0xfc, 0x1e, 0x00, 0xcc, 0xe6, 0x66, 0x00, 0x00, 0x00,
  0x00, 0xff, 0xbe, 0xf1, 0xe0, 0x7f, 0x83, 0xfc, 0x1c, 0x01, 0xcc, 0xe6, 0xe6, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x18, 0x61, 0xe0, 0xf7, 0x87, 0xbe, 0x3c, 0x00, 0xfc, 0xfe, 0xfe, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x1c, 0xe0, 0xff, 0xe7, 0xff, 0x1f, 0xf8, 0x00, 0x60, 0x78, 0x78, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x07, 0xc0, 0x3f, 0x81, 0xfc, 0x0f, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3b, 0xf7, 0xf7, 0xf3, 0xf9, 0xf8, 0xf8, 0xfc, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x11, 0xf1, 0xf1, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x7e, 0xfc, 0x50, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};


void setup()  {
  //Sensor Input
  pinMode(SENSOR_L, INPUT_PULLUP);
  pinMode(SENSOR_R, INPUT_PULLUP);

  //Interface baud rate zur seriellen Ausgabe über den seriellen Monitor (Rechnerbildschirm)
  Serial.begin(9600);

  //Display Inizialisierung
  initialize_display();
}

//-----------------------------------
void Delay_no_Block(unsigned long ms)
//-----------------------------------
// Beim normalen delay() werden die interrupts blockiert
// => Die Serielle schnittstelle kann nicht abgearbeitet werden.
{
  unsigned long Ende = millis() + ms;         // Aktuelle Zeit + Wartezeit (ms)
  while (millis() < Ende)
  {
  }
}

//-----------------------------------------
void Check_Sensor(int Sensor1, int Sensor2)
//-----------------------------------------
{
  // ToDo: Timeout

  //warten auf das Signal vom ersten Sensor
  if (digitalRead(Sensor1) == LOW)            // Zug erkannt
  {
    // Erster  Sensor wurde aktiviert, start time = system time (ms)
    starttime = millis();
    underprogress_display();
    //warten auf das Signal vom zweiten Sensor
    while (digitalRead(Sensor2) == HIGH) {}   // Warten solange wie kein Zug erkannt wird

    //Berechnung der Zeitdifferenz in ms
    deltatime = (millis() - starttime);

//    deltatime++;


    refresh_display();

    // Warten bis Zug ausgefahren ist
    freetime_counter = 0;
    while (freetime_counter < 20)             // Luecke zwischen den Wagons muss laenger als 20*100ms sein (2 Sek)
    {
      freetime_counter++;

      if (digitalRead(Sensor2) == LOW)        // Sensor dunkel
      {
        freetime_counter = 0;                 //Zeit zurücksetzen für neuen Start
      }
      Delay_no_Block(100);                    //alt: delay (100);
    }
    state_ready_display();
  }
}

//---------
void loop()
//---------
{
  #if ITRAIN_TEST >= 2
    iTrain_Test();
  #endif

  Check_Sensor(SENSOR_L, SENSOR_R);
  Check_Sensor(SENSOR_R, SENSOR_L);
} //Ende der LOOP-Schleife

#if ITRAIN_TEST >= 2
  //----------------
  void iTrain_Test()
  //----------------
  {
    display.clearDisplay();
    display.setTextSize(1);
    display.println("iTrainTest");
    display.println("Hallo");
    display.display();
    unsigned long NextSend = 0;
    
    long deltatime = 1000;
    while (1)
        {
        if (millis() > NextSend)
           {
           NextSend = millis() + 1000;

           float Speed = scale * distance / deltatime * 3.6;
           // Augabe fuer Speed Cat
           // Umrechnung:
           //  100 Impulse = 84 km/h bei 1:87
           //  => 1 km/h = 1000/843
           char txt[11];
           int imp = round(1.18694362 * round(Speed));
           sprintf(txt, "*%04i;V3.0%%", imp);
           Serial.println(txt);  //"*0100;V3.0%"
           }
        if (Serial.available())
           {
           char c = Serial.read();
            display.print(c);
            display.display();
           }
        }
  }
#endif

//--------------------
void refresh_display()
//--------------------
//Ausgaben für den seriellen Monitor und das OLED-Display
{
  float Speed = scale * distance / deltatime * 3.6;
  //Speed = deltatime;
#if ITRAIN_TEST == 0
  //Ausgabe für den seriellen Monitor (Rechnerbildschirm)
  Serial.println("Messung gestartet");
  Serial.println("");
  Serial.println("Zeitunterschied:");
  Serial.print(deltatime);
  Serial.println(" ms");
  Serial.println("---------");
  Serial.println("Geschwindigkeit:");
  Serial.print(distance / deltatime);
  Serial.println(" m/s (echt)");
  Serial.print(Speed, 2);
  Serial.print(" km/h (1/");
  Serial.print(scale);
  Serial.println(")");
#endif // ITRAIN_TEST

//--------------------
  // Augabe fuer Speed Cat
  // Umrechnung:
  //  100 Impulse = 84 km/h bei 1:87
  //  => 1 km/h = 1000/843
  char txt[11];
  int imp = round(1.15*1.18694362 * round(Speed));    //Faktor 1.15 für Korrekturumrechnung iTrain
  sprintf(txt, "*%04i;V3.0%%", imp);
  Serial.println(txt);  //"*0100;V3.0%"
  Serial.println(txt);  //"*0100;V3.0%" // SpeedCat zeigt erst dann an wenn weitere 2 Zeichen kommen ;-(

 //--------------------
  //Ausgabe auf dem OLED-Display
  display.clearDisplay();
  display.setTextSize(5);
  display.setCursor(0, 28);
  dtostrf(Speed,  3, 1, txt); // 99.2
  char LastDigit = txt[strlen(txt) - 1]; // Letzen Buchstaben kopieren
  *strchr(txt, '.') = '\0';              // Suche '.' ersetze mit Ende String Zeichen
  for(int i=strlen(txt); i<3; i++) display.print(' '); // Führende Leerzeichen falls txt kürzer als 3 Zeichen
  display.print(txt);
  display.setTextSize(2);
  display.setCursor(89, 50);
  display.print(LastDigit);
  display.setTextSize(1);
  display.setCursor(128 - 4 * 6, 56);
  display.print("km/h");
  sprintf(txt, "1/%i", scale);
  display.setCursor(128 - strlen(txt) * 6, 38);
  display.print(txt);
  display.drawFastHLine(0, 20, 128, WHITE);
  display.setTextSize(2);
  display.setCursor(0, 0);
  dtostrf(deltatime * 0.001, 4, 2, txt);  // 1.23 aber 17.56 wenn länger
  if (strlen(txt) > 4) dtostrf(deltatime * 0.001, 4, 1, txt);  //  17.5
  display.print(txt);
  display.setTextSize(1);
  display.setCursor(50, 7);
  display.print("s");
  display.drawFastVLine(63, 0, 19, WHITE);
  display.setTextSize(2);
  display.setCursor(71, 00);
  display.print(distance / deltatime);
  display.setTextSize(1);
  display.setCursor(121, -2);
  display.print("m");
  display.drawFastHLine(121, 7, 5, WHITE);
  display.setCursor(121, 8);
  display.print("s");
  display.display();
}

void state_ready_display() {
  display.setTextSize(1);
  display.setCursor(128 - 5 * 6, 25);
  display.print("Ready");
  display.display();
}

void underprogress_display() {

  display.clearDisplay();
  display.setTextSize(2);
  display.setCursor(0, 20);
  display.println("");
  display.print("gestartet");
  display.setTextSize(2);
  display.setCursor(5, 0);
  display.print("Messung");
  display.display();
}

void initialize_display() {
 #if OLED == 96
  display.begin(SSD1306_SWITCHCAPVCC, 0x3c);       // inizialisiert 0.96" OLED-Display (Adresse für original Adafruit Produkte ist 0x3D)
#else
  display.begin(SH1106_SWITCHCAPVCC, 0x3C);      // inizialisiert 1.3" OLED display (Adresse für original Adafruit Produkte ist 0x3D)
#endif
  display.display();
  display.clearDisplay();
  display.drawBitmap(0, 0, logo, 128, 64, WHITE);  //Logo wird gezeichnet und platziert
  display.display();
  display.setTextColor(INVERSE);  //Anzeige wird invertiert


  Delay_no_Block(600);

  // Daten des Erstellers
  display.clearDisplay();
  display.setTextSize(1);
  display.setCursor(0, 0);
  display.println("Idee von");
  display.println("N-MODELLBAHN");
  display.println("");
  display.println("verbessert durch");
  display.println("Juergen (fromue)");
  display.println("und Hardi");
  display.print("Vers. 1.6/ Nov. 2022");
  display.display();

  Delay_no_Block(3000);       //ich möchte das es zu Beginn gelesen wird :-)

  // 1. Bereitschaft der Software zur Messung
  display.clearDisplay();
  display.setTextSize(3);
  display.setCursor(7, 28);
  display.print(" Start ");
  display.drawFastHLine(0, 20, 128, WHITE);
  display.setTextSize(2);
  display.setCursor(37, 0);
  display.print("1/");
  display.setCursor(60, 0);
  display.print(scale);
  display.display();

  Serial.println("<<< START >>>");    //erste serielle Ausgabe der Messbereitschaft für Rechnerbildschirm nach Arduinostart
}
